<?php

namespace Drupal\Tests\layout_builder_ipe\FunctionalJavascript;

use Drupal\block_content\Entity\BlockContentType;
use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\Tests\contextual\FunctionalJavascript\ContextualLinkClickTrait;
use Drupal\Tests\layout_builder\FunctionalJavascript\InlineBlockTestBase;
use Drupal\Tests\system\Traits\OffCanvasTestTrait;

/**
 * Tests the Layout Builder IPE UI.
 *
 * @group layout_builder_ipe
 */
class LayoutBuilderIpeUiTest extends WebDriverTestBase {

  use ContextualLinkClickTrait;
  use OffCanvasTestTrait;

  const BUNDLE_LB = 'bundle_layout_builder';
  const BUNDLE_NO_LB = 'bundle_no_layout_builder';

  /**
   * Path prefix for the field UI for the test bundle.
   *
   * @var string
   */
  const FIELD_UI_PREFIX_LB = 'admin/structure/types/manage/' . self::BUNDLE_LB;

  /**
   * Path prefix for the field UI for the test bundle.
   *
   * @var string
   */
  const FIELD_UI_PREFIX_NO_LB = 'admin/structure/types/manage/' . self::BUNDLE_NO_LB;

  protected static $modules = [
    'layout_builder_ipe',
    'node',
    'field_ui',
    'contextual',
    'block',
    'block_content',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'starterkit_theme';

  /**
   * Node using Layout Builder.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $node_lb;

  /**
   * Node not using Layout Builder.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $node_no_lb;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    // Create a user with sufficient permissions to setup Layout Builder.
    $this->drupalLogin($this->drupalCreateUser([
      'configure any layout',
      'create and edit custom blocks',
      'administer node display',
      'administer node fields',
      'access contextual links',
      'administer layout builder ipe',
    ]));

    // Create content types.
    $this->createContentType(['type' => self::BUNDLE_LB]);
    $this->createContentType(['type' => self::BUNDLE_NO_LB]);

    // Enable layout builder for the first bundle.
    $this->drupalGet(static::FIELD_UI_PREFIX_LB . '/display/default');

    $page = $this->getSession()->getPage();
    $page->find('css', '[name="layout[enabled]"]')->check();
    $page->find('css', '[name="layout[allow_custom]"]')->check();
    $page->find('css', '[name="layout[layout_builder_ipe]"]')->check();
    $page->find('css', '[value="Save"]')->click();
    $this->assertSession()->pageTextContains('Your settings have been saved.');
    $this->drupalGet(static::FIELD_UI_PREFIX_LB . '/display/default');

    // Disable layout builder for the second bundle.
    $this->drupalGet(static::FIELD_UI_PREFIX_NO_LB . '/display/default');

    $page = $this->getSession()->getPage();
    $page->find('css', '[name="layout[enabled]"]')->uncheck();
    $page->find('css', '[name="layout[allow_custom]"]')->uncheck();
    $page->find('css', '[name="layout[layout_builder_ipe]"]')->uncheck();
    $page->find('css', '[value="Save"]')->click();
    $this->assertSession()->pageTextContains('Your settings have been saved.');
    $this->drupalGet(static::FIELD_UI_PREFIX_NO_LB . '/display/default');

    // Create a node for each bundle.
    $this->node_lb = $this->createNode([
      'type' => self::BUNDLE_LB,
      'title' => "Node LB title (nid 1)",
    ]);
    $this->node_no_lb = $this->createNode([
      'type' => self::BUNDLE_NO_LB,
      'title' => "Node NO LB title (nid 2)",
    ]);

    $this->drupalLogout();
  }

  /**
   * Tests disabling the layout page.
   */
  public function testDisableLayoutPage() {
    $user = $this->getLbAdminUser();

    // Disable the layout page.
    $this->setIpeSetting('disable_layout_page', TRUE, $user);

    // Try to open the layout page.
    $this->drupalGet($this->node_lb->toUrl()->toString() . '/layout');
    $assert_session = $this->assertSession();
    $assert_session->pageTextContains('You are not authorized to access this page.');

    // Disable the setting.
    $this->setIpeSetting('disable_layout_page', FALSE, $user);

    // Try to open the layout page again.
    $this->drupalGet($this->node_lb->toUrl()->toString() . '/layout');
    $assert_session->pageTextNotContains('You are not authorized to access this page.');
  }

  /**
   * Test the customize workflow for a non-overridden node.
   */
  public function testCustomizeWorkflow() {

    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // Save a layout without changes.
    $page->pressButton('Save layout');
    $assert_session->assertNoElementAfterWait('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->pageTextContains('The layout override has been saved.');

    // Customize again to discard.
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElement('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->buttonExists('Discard changes');
    $page->pressButton('Discard changes');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->responseContains('Are you sure you want to discard your layout changes?');
    $assert_session->buttonExists('Confirm');
    $page->pressButton('Confirm');
    $assert_session->assertNoElementAfterWait('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->pageTextContains('The changes to the layout have been discarded.');
  }

  /**
   * Tests that custom blocks can be added.
   */
  public function testOverrideLayout() {

    // Add a new block content bundle to the editorial workflow.
    BlockContentType::create([
      'id' => 'basic',
      'label' => 'Basic',
      'revision' => 1,
    ])->save();
    block_content_add_body_field('basic');

    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    $block_title = 'The block title';
    $block_body = 'The block body';

    // Customize.
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElement('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Add new block.
    $this->openAddBlockForm('Create custom block');

    $textarea = $assert_session->waitForElement('css', '[name="settings[block_form][body][0][value]"]');
    $this->assertNotEmpty($textarea);
    $assert_session->fieldValueEquals('Title', '');
    $page->findField('Title')->setValue($block_title);
    $textarea->setValue($block_body);
    $page->pressButton('Add block');
    $this->assertDialogClosedAndTextVisible($block_body, InlineBlockTestBase::INLINE_BLOCK_LOCATOR);

    // Save a layout with changes.
    $page->pressButton('Save layout');
    $assert_session->assertNoElementAfterWait('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->pageTextContains('The layout override has been saved.');

    // Confirm the new block is there.
    $assert_session->pageTextContains($block_title);
    $assert_session->pageTextContains($block_body);
  }

  /**
   * Tests that IPE interface is not displayed if user has no IPE permission.
   */
  public function testNoAccessNoIpePermission() {
    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session = $this->assertSession();
    $assert_session->elementNotExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementNotExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkNotExists('Customize');
  }

  /**
   * Tests that IPE interface is not displayed if disabled for the view mode.
   */
  public function testNoAccessViewModeDisabled() {
    $this->drupalLogin($this->drupalCreateUser([
      'configure any layout',
      'create and edit custom blocks',
      'administer node display',
      'administer node fields',
      'access contextual links',
      'administer layout builder ipe',
    ]));

    // Enable layout builder.
    $this->drupalGet(static::FIELD_UI_PREFIX_LB . '/display/default');

    $page = $this->getSession()->getPage();
    $page->find('css', '[name="layout[enabled]"]')->check();
    $page->find('css', '[name="layout[allow_custom]"]')->check();
    $page->find('css', '[name="layout[layout_builder_ipe]"]')->uncheck();
    $page->find('css', '[value="Save"]')->click();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session = $this->assertSession();
    $assert_session->elementNotExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementNotExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkNotExists('Customize');
  }

  /**
   * Tests that IPE interface is not displayed if user has no edit permission.
   */
  public function testNoAccessNoEditPermission() {
    $this->drupalLogin($this->drupalCreateUser([
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session = $this->assertSession();
    $assert_session->elementNotExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementNotExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkNotExists('Customize');
  }

  /**
   * Tests that IPE interface is not displayed if the node doesn't use LB.
   */
  public function testNoAccessNoLayoutBuilder() {
    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_NO_LB . ' content',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    // Go to node view page.
    $this->drupalGet($this->node_no_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session = $this->assertSession();
    $assert_session->elementNotExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementNotExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkNotExists('Customize');
  }

  /**
   * Tests that IPE correctly shows error messages for concurrent edits.
   */
  public function testConcurrentEdits() {
    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Now, without discarding or saving, navigate to the edit form and save that.
    $this->drupalGet($this->node_lb->toUrl('edit-form')->toString());
    $page->pressButton('Save');

    // Customize again and try to save. We should see an error that the page
    // can't be saved due to modifications done in the meantime.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // Save a layout without changes.
    $page->pressButton('Save layout');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->responseContains('The content has either been modified by another user, or you have already submitted modifications. As a result, your changes cannot be saved.');
  }

  /**
   * Tests that IPE correctly shows error messages for concurrent edits.
   */
  public function testConcurrentEditsEnhanced() {
    // Enable enhanced entity changed checking.
    $this->setIpeSetting('override_entity_changed_constraint', TRUE);

    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Now, without discarding or saving, navigate to the edit form and save that.
    $this->drupalGet($this->node_lb->toUrl('edit-form')->toString());
    $page->pressButton('Save');

    // Customize again and try to save. This should work.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // Save a layout without changes.
    $page->pressButton('Save layout');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->addressEquals($this->node_lb->toUrl()->toString());
  }

  /**
   * Tests that IPE uses shared layout sessions per user.
   */
  public function testSharedLayoutSessionPerUser() {
    // Enable enhanced entity changed checking.
    $this->setIpeSetting('non_concurrent_editing', FALSE);

    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->pageTextNotContains('Powered by Drupal');

    // Add a new block.
    $this->openAddBlockForm('Powered by Drupal');

    $page->fillField('settings[label]', 'This is the label');
    $page->checkField('settings[label_display]');

    // Save the new block, and ensure it is displayed on the page.
    $page->pressButton('Add block');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->assertNoElementAfterWait('css', '#drupal-off-canvas');
    $assert_session->pageTextContains('Powered by Drupal');
    $assert_session->pageTextContains('This is the label');

    // Now login as a new user with the same permissions.
    $this->drupalLogin($this->drupalCreateUser([
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ]));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // And confirm that the changes made by the previous user are visible
    // to this one.
    $assert_session->pageTextContains('Powered by Drupal');
    $assert_session->pageTextContains('This is the label');
  }

  /**
   * Tests that IPE uses separate layout sessions per user.
   */
  public function testSeparateLayoutSessionPerUser() {
    // Enable enhanced entity changed checking.
    $this->setIpeSetting('non_concurrent_editing', TRUE);

    $permissions = [
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ];
    $this->drupalLogin($this->drupalCreateUser($permissions));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');
    $assert_session->pageTextNotContains('Powered by Drupal');

    // Add a new block.
    $this->openAddBlockForm('Powered by Drupal');

    $page->fillField('settings[label]', 'This is the label');
    $page->checkField('settings[label_display]');

    // Save the new block, and ensure it is displayed on the page.
    $page->pressButton('Add block');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->assertNoElementAfterWait('css', '#drupal-off-canvas');
    $assert_session->pageTextContains('Powered by Drupal');
    $assert_session->pageTextContains('This is the label');

    // Now login as a new user with the same permissions.
    $this->drupalLogin($this->drupalCreateUser($permissions));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // And confirm that the changes made by the previous user are not visible
    // to this one.
    $assert_session->pageTextNotContains('Powered by Drupal');
    $assert_session->pageTextNotContains('This is the label');
  }

  /**
   * Tests that IPE uses separate layout sessions per user.
   */
  public function testLayoutLock() {
    // Enable locking.
    $this->setIpeSetting('non_concurrent_editing', FALSE);
    $this->setIpeSetting('lock_layout', TRUE);

    $permissions = [
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
    ];
    $user_1 = $this->drupalCreateUser($permissions);
    $this->drupalLogin($user_1);

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');
    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // We need to save at least once to create an override section storage.
    $page = $this->getSession()->getPage();
    $page->pressButton('Save layout');
    $assert_session->assertWaitOnAjaxRequest();

    // Now click customize again to lock this layout.
    $page = $this->getSession()->getPage();
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');

    // Now login as a new user with the same permissions.
    $this->drupalLogin($this->drupalCreateUser($permissions));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#drupal-modal');
    $assert_session->pageTextContains('This page is currently being edited by ' . $user_1->getDisplayName() . ' and is therefore locked from editing by others');
    $assert_session->buttonExists('Ok');
    $assert_session->buttonNotExists('Break lock');
  }

  /**
   * Tests that IPE uses separate layout sessions per user.
   */
  public function testLayoutLockWithBreak() {
    // Enable locking.
    $this->setIpeSetting('non_concurrent_editing', FALSE);
    $this->setIpeSetting('lock_layout', TRUE);

    $permissions = [
      'edit any ' . self::BUNDLE_LB . ' content',
      'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
      'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
      'create and edit custom blocks',
      'access contextual links',
      'layout builder ipe break locks',
    ];
    $user_1 = $this->drupalCreateUser($permissions);
    $this->drupalLogin($user_1);

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');
    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // We need to save at least once to create an override section storage.
    $page = $this->getSession()->getPage();
    $page->pressButton('Save layout');
    $assert_session->assertWaitOnAjaxRequest();

    // Now click customize again to lock this layout.
    $page = $this->getSession()->getPage();
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#layout-builder-ipe-wrapper.edit-layout');
    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Now login as a new user with the same permissions.
    $this->drupalLogin($this->drupalCreateUser($permissions));

    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    // Go to node view page.
    $this->drupalGet($this->node_lb->toUrl()->toString());

    // Click the customize link to open an edit session.
    $assert_session->elementExists('css', '#layout-builder-ipe-wrapper');
    $assert_session->elementExists('css', '.layout-builder-ipe-actions');
    $assert_session->linkExists('Customize');
    $page->clickLink('Customize');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '[role="dialog"] #drupal-modal');
    $assert_session->elementExists('css', '[role="dialog"] #drupal-modal');
    $assert_session->pageTextContains('This page is currently being edited by ' . $user_1->getDisplayName() . ' and is therefore locked from editing by others');
    $assert_session->pageTextContains('You can break the look to start editing this page');
    $assert_session->buttonExists('Ok');
    $assert_session->buttonExists('Break lock');

    // Now break the lock and confirm that it's working.
    $page->pressButton('Break lock');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->assertNoElementAfterWait('css', '.ui-dialog.locked-message');
    // Check that the form and the three main buttons are there.
    $assert_session->elementExists('css', 'form.layout-builder-form');
    $assert_session->buttonExists('Revert to defaults');
    $assert_session->buttonExists('Discard changes');
    $assert_session->buttonExists('Save layout');
  }

  /**
   * Set a global IPE setting.
   *
   * @param string $settings_key
   *   The key of the setting.
   */
  private function setIpeSetting($settings_key, $state, $user = NULL) {
    if (!$user) {
      $user = $this->getLbAdminUser();
    }
    // Confirm that the setting to disable the layout page is checked.
    $this->drupalGet('admin/config/user-interface/layout-builder-ipe');
    $page = $this->getSession()->getPage();
    $assert_session = $this->assertSession();
    if ($state) {
      $page->find('css', '[name="' . $settings_key . '"]')->check();
    }
    else {
      $page->find('css', '[name="' . $settings_key . '"]')->uncheck();
    }
    $page->find('css', '[value="Save configuration"]')->click();
    $assert_session->responseContains('The configuration options have been saved.');

    if ($state) {
      $assert_session->checkboxChecked($settings_key);
    }
    else {
      $assert_session->checkboxNotChecked($settings_key);
    }
  }

  /**
   * Get a user object with a full set of Layout Builder permissions.
   *
   * @return \Drupal\Core\Session\AccountInterface
   *   A user account object.
   */
  private function getLbAdminUser() {
    $user = &drupal_static(__FUNCTION__);
    if (!$user) {
      $user = $this->drupalLogin($this->drupalCreateUser([
        'edit any ' . self::BUNDLE_LB . ' content',
        'configure editable ' . self::BUNDLE_LB . ' node layout overrides',
        'use layout builder ipe on editable ' . self::BUNDLE_LB . ' node layout overrides',
        'create and edit custom blocks',
        'administer node display',
        'administer node fields',
        'access contextual links',
        'administer layout builder ipe',
      ]));
    }
    return $user;
  }

  /**
   * Opens the add block form in the off-canvas dialog.
   *
   * @param string $block_title
   *   The block title which will be the link text.
   */
  private function openAddBlockForm($block_title) {
    $assert_session = $this->assertSession();
    $assert_session->linkExists('Add block');
    $this->clickLink('Add block');
    $assert_session->assertWaitOnAjaxRequest();
    $this->assertNotEmpty($assert_session->waitForElementVisible('named', ['link', $block_title]));
    $this->clickLink($block_title);
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->waitForElementVisible('css', '#drupal-off-canvas');
  }

  /**
   * Asserts that the dialog closes and the new text appears on the main canvas.
   *
   * @param string $text
   *   The text.
   * @param string|null $css_locator
   *   The css locator to use inside the main canvas if any.
   */
  protected function assertDialogClosedAndTextVisible($text, $css_locator = NULL) {
    $assert_session = $this->assertSession();
    $assert_session->assertNoElementAfterWait('css', '#drupal-off-canvas');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->elementNotExists('css', '#drupal-off-canvas');
    if ($css_locator) {
      $this->assertNotEmpty($assert_session->waitForElementVisible('css', ".dialog-off-canvas-main-canvas $css_locator:contains('$text')"));
    }
    else {
      $this->assertNotEmpty($assert_session->waitForElementVisible('css', ".dialog-off-canvas-main-canvas:contains('$text')"));
    }
  }

}
